/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

/*******************************************************************************
/
/   File:      PortablePoint.h
/
/   Description:    Portable version of Be's PortablePoint class
/
*******************************************************************************/

#ifndef PORTABLE_POINT_H
#define PORTABLE_POINT_H

#include <stdio.h>
#include "support/PortableFlattenable.h"
#include "support/PortableTypeConstants.h"

/*----------------------------------------------------------------------*/
/*----- PortablePoint class --------------------------------------------*/

/** A portable version of Be's BPoint class. */
class MUSCLE_CLASS_DECLARATIONS PortablePoint : public PortableFlattenable
{
public:
   /** Exposed member variable representing the X coordinate of the point. */
   float x;

   /** Exposed member variable representing the X coordinate of the point. */
   float y;

   /** Default constructor, sets the point to be (0, 0) */
   PortablePoint() {x = y = 0.0f;}

   /** Constructor where you specify the initial value of the point 
    *  @param X Initial X position
    *  @param Y Initial Y position
    */
   PortablePoint(float X, float Y) {Set(X, Y);}

   /** Copy Constructor. */
   PortablePoint(const PortablePoint & rhs) : PortableFlattenable() {*this = rhs;}

   /** Destructor */
   virtual ~PortablePoint() {/* empty */}
 
   /** Sets a new value for the point.
    *  @param X The new X value 
    *  @param Y The new Y value 
    */
   void Set(float X, float Y) {x = X; y = Y;}

   /** If the point is outside the rectangle specified by the two arguments,
    *  it will be moved horizontally and/or vertically until it falls inside the rectangle.
    *  @param topLeft Minimum values acceptable for X and Y
    *  @param bottomRight Maximum values acceptable for X and Y
    */
   void ConstrainTo(PortablePoint topLeft, PortablePoint bottomRight)
   {
      if (x < topLeft.x) x = topLeft.x;
      if (y < topLeft.y) y = topLeft.y;
      if (x > bottomRight.x) x = bottomRight.x;
      if (y > bottomRight.y) y = bottomRight.y;
   }

   /** Print debug information about the point to stdout */
   void PrintToStream() const
   {
      printf("PortablePoint: %f %f\n", x, y);
   }
      
   /** Returns another PortablePoint whose values are the sum of this point's values and (p)'s values */
   PortablePoint operator+(const PortablePoint&p) const
   {
      return PortablePoint(x+p.x, y+p.y);
   }

   /** Returns another PortablePoint whose values are the difference of this point's values and (p)'s values */
   PortablePoint operator-(const PortablePoint&p) const
   {
      return PortablePoint(x-p.x, y-p.y);
   }

   /** Adds (p)'s values to this point's values */
   PortablePoint& operator+=(const PortablePoint&p)
   {
      x += p.x;
      y += p.y;
      return *this;
   }

   /** Subtracts (p)'s values from this point's values */
   PortablePoint& operator-=(const PortablePoint&p)
   {
      x -= p.x;
      y -= p.y;
      return *this;
   }

   /** Returns true iff (p)'s values don't match this point's values */
   bool operator!=(const PortablePoint&p) const
   {
      return ((x != p.x)||(y != p.y));
   }

   /** Returns true iff (p)'s values match this point's values */
   bool operator==(const PortablePoint&p) const
   {
      return ((x == p.x)&&(y == p.y));
   }

   /** Part of the PortableFlattenable interface:  Returns true */
   virtual bool IsFixedSize() const {return true;} 

   /** Part of the PortableFlattenable interface:  Returns B_POINT_TYPE */
   virtual type_code TypeCode() const {return B_POINT_TYPE;}

   /** Part of the PortableFlattenable interface:  2*sizeof(float) */
   virtual size_t FlattenedSize() const {return 2 * sizeof(float);}

   /** Copies this point into an endian-neutral flattened buffer.
    *  @param buffer Points to an array of at least FlattenedSize() bytes.
    */
   virtual void Flatten(uint8 * buffer) const 
   {
      float * buf = (float *) buffer;
      buf[0] = B_HOST_TO_LENDIAN_FLOAT(x);
      buf[1] = B_HOST_TO_LENDIAN_FLOAT(y);
   }

   /** Restores this point from an endian-neutral flattened buffer.
    *  @param buffer Points to an array of (size) bytes
    *  @param size The number of bytes (buffer) points to (should be at least FlattenedSize())
    *  @return B_NO_ERROR on success, B_ERROR on failure (size was too small)
    */
   virtual status_t Unflatten(const uint8 * buffer, size_t size) 
   {
      if (size >= FlattenedSize())
      {
         float * buf = (float *) buffer;
         x = B_LENDIAN_TO_HOST_FLOAT(buf[0]);
         y = B_LENDIAN_TO_HOST_FLOAT(buf[1]);
         return B_NO_ERROR;
      }
      else return B_ERROR;
   }
};

#endif 
