/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef _PORTABLE_QUEUE_H
#define _PORTABLE_QUEUE_H

#include "support/PortableTypeDefs.h"
#include "support/PortableMacros.h"

#define SMALL_QUEUE_SIZE 3

/** This class implements a templated double-ended queue data structure.
 *  Adding or removing items from the head or tail of a PortableQueue is (on average)
 *  an O(1) operation.  A PortableQueue also makes for a nice Vector, if that's all you need.
 */
template <class ItemType> class MUSCLE_CLASS_DECLARATIONS PortableQueue
{
public:
   /** Constructor.
    *  @param initialSize Specifies how many slots to pre-allocate.  Defaults to (SMALL_QUEUE_SIZE).
    */
   PortableQueue(uint32 initialSize = SMALL_QUEUE_SIZE);

   /** Copy constructor. */
   PortableQueue(const PortableQueue& copyMe);

   /** Destructor. */
   virtual ~PortableQueue();

   /** Assigment operator. */
   PortableQueue &operator=(const PortableQueue &from);

   /** Equality operator.  Queues are equal if they are the same length, and
     * every nth item in this queue is == to the corresponding item in (rhs). */
   bool operator==(const PortableQueue &rhs) const;

   /** Returns the negation of the equality operator */
   bool operator!=(const PortableQueue &rhs) const {return !(*this == rhs);}

   /** Appends (item) to the end of the queue.  Queue size grows by one.
    *  @param item The item to append. 
    *  @return True on success, false on failure (out of memory)
    */
   bool AddTail(const ItemType &item);

   /** Appends all items in (queue) to the end of our queue.  Queue size
    *  grows by (queue.CountItems()).
    *  For example:
    *    PortableQueue a;   // contains 1, 2, 3, 4
    *    PortableQueue b;   // contains 5, 6, 7, 8
    *    a.AddHead(b);      // a now contains 1, 2, 3, 4, 5, 6, 7, 8
    *  @param item The queue to append to our queue.
    *  @return True on success, false on failure (out of memory)
    */
   bool AddTail(const PortableQueue<ItemType> & queue);

   /** Prepends (item) to the head of the queue.  Queue size grows by one.
    *  @param item The item to prepend. 
    *  @return True on success, false on failure (out of memory)
    */
   bool AddHead(const ItemType &item);

   /** Concatenates (queue) to the head of our queue.
    *  Our queue size grows by (queue.CountItems()).
    *  For example:
    *    PortableQueue a;   // contains 1, 2, 3, 4
    *    PortableQueue b;   // contains 5, 6, 7, 8
    *    a.AddHead(b);      // a now contains 5, 6, 7, 8, 1, 2, 3, 4
    *  @param item The queue to prepend to our queue.
    *  @return True on success, false on failure (out of memory)
    */
   bool AddHead(const PortableQueue<ItemType> & queue);

   /** Removes the item at the head of the queue and places it into (returnItem).
    *  @param returnItem On success, the removed item is copied into this object.
    *  @return true on success, false if the queue was empty
    */
   bool RemoveHead(ItemType & returnItem);

   /** Removes the item at the head of the queue.
    *  @return True on success, false if the queue was empty.
    */
   bool RemoveHead();

   /** Removes the item at the tail of the queue and places it into (returnItem).
    *  @param returnItem On success, the removed item is copied into this object.
    *  @return true on success, false if the queue was empty
    */
   bool RemoveTail(ItemType & returnItem);

   /** Removes the item at the tail of the queue.
    *  @return True on success, false if the queue was empty.
    */
   bool RemoveTail();

   /** Removes the item at the (index)'th position in the queue.
    *  @param index Which item to remove--can range from zero 
    *               (head of the queue) to CountItems()-1 (tail of the queue).
    *  @return true on success, false on failure (i.e. bad index)
    *  Note that this method is somewhat inefficient for indices that
    *  aren't at the head or tail of the queue (i.e. O(n) time)
    */
   bool RemoveItemAt(uint32 index);

   /** Removes the item at the (index)'th position in the queue, and copies it into (returnItem).
    *  @param index Which item to remove--can range from zero 
    *               (head of the queue) to (CountItems()-1) (tail of the queue).
    *  @param returnItem On success, the removed item is copied into this object.
    *  @return true on success, false on failure (i.e. bad index)
    */
   bool RemoveItemAt(uint32 index, ItemType & returnItem);

   /** Copies the (index)'th item into (returnItem).
    *  @param index Which item to get--can range from zero 
    *               (head of the queue) to (CountItems()-1) (tail of the queue).
    *  @param returnItem On success, the retrieved item is copied into this object.
    *  @return true on success, false on failure (e.g. bad index)
    */
   bool GetItemAt(uint32 index, ItemType & returnItem) const;

   /** Replaces the (index)'th item in the queue with (newItem).
    *  @param index Which item to replace--can range from zero 
    *               (head of the queue) to (CountItems()-1) (tail of the queue).
    *  @param newItem The item to place into the queue at the (index)'th position.
    *  @return true on success, false on failure (e.g. bad index)
    */
   bool ReplaceItemAt(uint32 index, const ItemType & newItem);
 
   /** Inserts (item) into the (nth) slot in the array.  InsertItemAt(0)
    *  is the same as AddHead(item), InsertItemAt(CountItems()) is the same
    *  as AddTail(item).  Other positions will involve an O(n) shifting of contents.
    *  @param index The position at which to insert the new item.
    *  @param newItem The item to insert into the queue.
    *  @return true on success, false on failure (i.e. bad index).
    */
   bool InsertItemAt(uint32 index, const ItemType & newItem);
   
   /** Removes all items from the queue. */
   void MakeEmpty();

   /** Returns the number of items in the queue.  (This number does not include pre-allocated space) */
   uint32 CountItems() const {return _itemCount;}

   /** Returns true iff their are no items in the queue. */
   bool IsEmpty() const {return (_itemCount == 0);}

   /** Returns the head item in the queue.  You must not call this when the queue is empty! */
   ItemType Head() const {return *GetItemPointer(0);}

   /** Returns the tail item in the queue.  You must not call this when the queue is empty! */
   ItemType Tail() const {return *GetItemPointer(_itemCount-1);}

   /** Returns a pointer to the first item in the queue, or NULL if the queue is empty */
   ItemType * HeadPointer() const {return (_itemCount > 0) ? GetItemPointer(0) : NULL;}

   /** Returns a pointer to the last item in the queue, or NULL if the queue is empty */
   ItemType * TailPointer() const {return (_itemCount > 0) ? GetItemPointer(_itemCount-1) : NULL;}

   /** Convenient array-style get operator (be sure to only use valid indices!) */
   ItemType operator [](uint32 Index) const; 

   /** Convenient array-style set operator (be sure to only use valid indices!) */
   ItemType & operator [](uint32 Index);

   /** Returns a pointer to an item in the array (i.e. no 
    *  copying of the item is done).  Included for efficiency;
    *  be careful with this, modifying the queue can invalidate 
    *  the returned pointer!
    *  @param index Index of the item to return a pointer to.  Must be a valid index.
    *  @return a pointer to the internally held item
    */
   ItemType * GetItemPointer(uint32 index) const;

   /** Makes sure there is enough space allocated for at least
    *  (numItems) items.  Returns true on success, false on
    *  failure (out of memory).  You only need to call this if
    *  you wish to minimize the number of data re-allocations done.
    *  @param numItems the minimum amount of items to pre-allocate space for in the PortableQueue.
    */
   bool EnsureSize(uint32 numItems);

   /** Returns the last index of the given (item), or -1 if (item) is
    *  not found in the list.  O(n) search time.
    *  @param item The item to look for.
    *  @return The index of (item), or -1 if no such item is present.
    */
   int32 IndexOf(const ItemType & item) const;

   /**
    *  Swaps the values of the two items at the given indices.  This operation
    *  will involve three copies of the held items.
    *  @param fromIndex First index.   0 < fromIndex < CountItems().
    *  @param toIndex.  Second index.  0 < toIndex   < CountItems().
    */
   void Swap(uint32 fromIndex, uint32 toIndex);

   /**
    *  Reverses the ordering of the items in the given subrange.
    *  (e.g. if the items were A,B,C,D,E, this would change them to E,D,C,B,A)
    *  @param from Index of the start of the subrange.  Defaults to zero.
    *  @param to Index of the next item after the end of the subrange.  If greater than
    *         the number of items currently in the queue, this value will be clipped
    *         to be equal to that number.  Defaults to the largest possible uint32.
    */
   void ReverseItemOrdering(uint32 from=0, uint32 to = ((uint32)-1)); 

   /**
    *  This is the signature of the type of callback function that you must pass 
    *  into the Sort() method.  This function should work like strcmp(), returning
    *  a negative value if (item1) is less than item2, or zero if the items are
    *  equal, or a positive value if (item1) is greater than item2.
    *  @param item1 The first item
    *  @param item2 The second item
    *  @param cookie A user-defined value that was passed in to the Sort() method.
    */
   typedef int (*ItemCompareFunc)(const ItemType & item1, const ItemType & item2, void * cookie);

   /**
    *  Does an in-place, stable sort on a subrange of the contents of this PortableQueue.
    *  (The sort algorithm is a smart, in-place merge sort).
    *  @param compareFunc A function that compares two items in this queue and returns
    *         a value indicating which is "larger".  (negative indicates
    *         that the second parameter is larger, positive indicate that the
    *         first is larger, and zero indicates that the two parameters are equal)
    *  @param from Index of the start of the subrange.  Defaults to zero.
    *  @param to Index of the next item after the end of the subrange.  
    *         If greater than the number of items currently in the queue, 
    *         the subrange will extend to the last item.  Defaults to the largest possible uint32.
    *  @param optCookie A user-defined value that will be passed to the (compareFunc).
    */
   void Sort(ItemCompareFunc compareFunc, uint32 from=0, uint32 to = ((uint32)-1), void * optCookie = NULL);

   /**
    *  Goes through the array and removes every item that is equal to (val).
    *  @param val the item to look for and remove
    *  @return The number of instances of (val) that were found and removed during this operation.
    */
   uint32 RemoveAllInstancesOf(const ItemType & val);

private:
   uint32 NextIndex(uint32 idx) const {return (idx >= _queueSize-1) ? 0 : idx+1;}
   uint32 PrevIndex(uint32 idx) const {return (idx <= 0) ? _queueSize-1 : idx-1;}

   // Translates a user-index into an index into the _queue array.
   uint32 InternalizeIndex(uint32 idx) const {return (_headIndex + idx) % _queueSize;}

   // Helper methods, used for sorting (stolen from http://www-ihm.lri.fr/~thomas/VisuTri/inplacestablesort.html)
   void Merge(ItemCompareFunc compareFunc, uint32 from, uint32 pivot, uint32 to, uint32 len1, uint32 len2, void * cookie);
   uint32 Lower(ItemCompareFunc compareFunc, uint32 from, uint32 to, const ItemType & val, void * cookie) const;
   uint32 Upper(ItemCompareFunc compareFunc, uint32 from, uint32 to, const ItemType & val, void * cookie) const;

   ItemType _smallQueue[SMALL_QUEUE_SIZE];  // small queues can be stored inline in this array
   ItemType * _queue;                       // points to _smallQueue, or to a dynamically alloc'd array
   uint32 _queueSize;  // number of slots in the _queue array
   uint32 _itemCount;  // number of valid items in the array
   int32 _headIndex;   // index of the first filled slot, or -1 
   int32 _tailIndex;   // index of the last filled slot, or -1
   const uint32 _initialSize;  // as specified in ctor

   // used to "reset" slots, which will free memory when using PortableRefs
#ifdef WIN32
   ItemType _blankItem;
#else
   const ItemType _blankItem;
#endif
};

template <class ItemType>
PortableQueue<ItemType>::PortableQueue(uint32 initialSize)
   : _queue(NULL), _queueSize(0), _itemCount(0), _initialSize(initialSize), _blankItem()
{
   // empty
}

template <class ItemType>
PortableQueue<ItemType>::PortableQueue(const PortableQueue& rhs)
   : _queue(NULL), _queueSize(0), _itemCount(0), _initialSize(rhs._initialSize), _blankItem()
{
   *this = rhs;
}

template <class ItemType>
bool
PortableQueue<ItemType>::operator ==(const PortableQueue& rhs) const
{
   if (this == &rhs) return true;
   if (CountItems() != rhs.CountItems()) return false;

   for (int i = CountItems()-1; i>=0; i--) if ((*this)[i] != rhs[i]) return false;

   return true;
}


template <class ItemType>
PortableQueue<ItemType> &
PortableQueue<ItemType>::operator =(const PortableQueue& rhs)
{
   if (this != &rhs)
   {
      MakeEmpty();
      uint32 numItems = rhs.CountItems();
      if (EnsureSize(numItems)) 
      {
         for (uint32 i=0; i<numItems; i++) 
         {
            if (AddTail(*rhs.GetItemPointer(i)) == false) break;  // trouble!?  Shouldn't happen...
         }
      }
   }
   return *this;
}

template <class ItemType>
ItemType &
PortableQueue<ItemType>::operator[](uint32 i)
{
   MASSERT(i<_itemCount, "Invalid index to PortableQueue::[]");
   return _queue[InternalizeIndex(i)];
}

template <class ItemType>
ItemType
PortableQueue<ItemType>::operator[](uint32 i) const
{
   MASSERT(i<_itemCount, "Invalid index to PortableQueue::[]");
   return _queue[InternalizeIndex(i)];
}             

template <class ItemType>
ItemType * 
PortableQueue<ItemType>::GetItemPointer(uint32 i) const
{
   return &_queue[InternalizeIndex(i)];
}

template <class ItemType>
PortableQueue<ItemType>::~PortableQueue()
{
   if (_queue != _smallQueue) delete [] _queue;
}


template <class ItemType>
bool 
PortableQueue<ItemType>::
AddTail(const ItemType &item)
{
   if (EnsureSize(_itemCount+1) == false) return false;
   if (_itemCount == 0) _headIndex = _tailIndex = 0;
                   else _tailIndex = NextIndex(_tailIndex);
   _queue[_tailIndex] = item;
   _itemCount++;
   return true;
}


template <class ItemType>
bool 
PortableQueue<ItemType>::
AddTail(const PortableQueue<ItemType> &queue)
{
   uint32 qSize = queue.CountItems();
   for (int i=0; i<qSize; i++) if (AddTail(queue[i]) == false) return false;
   return true;
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
AddHead(const ItemType &item)
{
   if (EnsureSize(_itemCount+1) == false) return false;
   if (_itemCount == 0) _headIndex = _tailIndex = 0;
                   else _headIndex = PrevIndex(_headIndex);
   _queue[_headIndex] = item;
   _itemCount++;
   return true;
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
AddHead(const PortableQueue<ItemType> &queue)
{
   for (int i=((int)queue.CountItems())-1; i>=0; i--) if (AddHead(queue[i]) == false) return false;
   return true;
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
RemoveHead(ItemType & returnItem)
{
   if (_itemCount == 0) return false;
   returnItem = _queue[_headIndex];
   return RemoveHead();
}

template <class ItemType>
bool
PortableQueue<ItemType>::
RemoveHead()
{
   if (_itemCount == 0) return false;
   int oldHeadIndex = _headIndex;
   _headIndex = NextIndex(_headIndex);
   _itemCount--;
   _queue[oldHeadIndex] = _blankItem;  // this must be done last, as queue state must be coherent when we do this
   return true;
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
RemoveTail(ItemType & returnItem)
{
   if (_itemCount == 0) return false;
   returnItem = _queue[_tailIndex];
   return RemoveTail();
}


template <class ItemType>
bool 
PortableQueue<ItemType>::
RemoveTail()
{
   if (_itemCount == 0) return false;
   int removedItemIndex = _tailIndex;
   _tailIndex = PrevIndex(_tailIndex);
   _itemCount--;
   _queue[removedItemIndex] = _blankItem;  // this must be done last, as queue state must be coherent when we do this
   return true;
}

template <class ItemType>
bool
PortableQueue<ItemType>::
GetItemAt(uint32 index, ItemType & returnItem) const
{
   if (index >= _itemCount) return false;
   returnItem = _queue[InternalizeIndex(index)];
   return true;
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
RemoveItemAt(uint32 index, ItemType & returnItem)
{
   if (index >= _itemCount) return false;
   returnItem = _queue[InternalizeIndex(index)];
   return RemoveItemAt(index);
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
RemoveItemAt(uint32 index)
{
   if (index >= _itemCount) return false;
   if (index == 0) return RemoveHead();  // this is cheaper!

   index = InternalizeIndex(index);
   while(index != (uint32) _tailIndex)
   {
      uint32 next = NextIndex(index);
      _queue[index] = _queue[next];
      index = next;
   }
   int oldIndex = _tailIndex;
   _tailIndex = PrevIndex(_tailIndex); 
   _itemCount--;
   _queue[oldIndex] = _blankItem;  // this must be done last, as queue state must be coherent when we do this
   return true; 
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
ReplaceItemAt(uint32 index, const ItemType & newItem)
{
   if (index >= _itemCount) return false;
   _queue[InternalizeIndex(index)] = newItem;
   return true;
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
InsertItemAt(uint32 index, const ItemType & newItem)
{
   // Simple cases
   if (index >  _itemCount) return false;
   if (index == _itemCount) return AddTail(newItem);
   if (index == 0)          return AddHead(newItem);

   // Harder case:  inserting into the middle of the array
   if (AddTail(newItem) == false) return false;  // just to allocate the extra slot
   int lo = (int) index;
   for (int i=((int)_itemCount)-1; i>lo; i--) ReplaceItemAt(i, *GetItemPointer(i-1));
   ReplaceItemAt(index, newItem);
   return true;
}

template <class ItemType>
void 
PortableQueue<ItemType>::
MakeEmpty()
{
   while(RemoveTail()) {/* empty */}
}

template <class ItemType>
bool 
PortableQueue<ItemType>::
EnsureSize(uint32 size)
{
   if ((_queue == NULL)||(_queueSize < size))
   {
      const uint32 sqLen = ARRAYITEMS(_smallQueue);
      uint32 newQLen = (size <= sqLen) ? sqLen : (size*2);  // if we're gonna do a new[], double it for later
      if (newQLen < _initialSize) newQLen = _initialSize;

      ItemType * newQueue = ((_queue == _smallQueue)||(newQLen > sqLen)) ? newnothrow ItemType[newQLen] : _smallQueue;
      if (newQueue == NULL) {WARN_OUT_OF_MEMORY; return false;}
      if (newQueue == _smallQueue) newQLen = sqLen;
      
      if (_itemCount > 0)
      {
         for (uint32 i=0; i<_itemCount; i++) (void) GetItemAt(i, newQueue[i]);
         _headIndex = 0;
         _tailIndex = _itemCount-1;
      }

      if (_queue == _smallQueue) for (uint32 i=0; i<sqLen; i++) _smallQueue[i] = _blankItem;
                            else delete [] _queue;

      _queue = newQueue;
      _queueSize = newQLen;
   }
   return true;
}

template <class ItemType>
int32 
PortableQueue<ItemType>::
IndexOf(const ItemType & item) const
{
   if (_queue) for (int i=((int)CountItems())-1; i>=0; i--) if (*GetItemPointer(i) == item) return i;
   return -1;
}


template <class ItemType>
void 
PortableQueue<ItemType>::
Swap(uint32 fromIndex, uint32 toIndex) 
{
   ItemType temp = *(GetItemPointer(fromIndex));
   ReplaceItemAt(fromIndex, *(GetItemPointer(toIndex)));
   ReplaceItemAt(toIndex,   temp);
}

template <class ItemType>
void 
PortableQueue<ItemType>::
Sort(ItemCompareFunc compareFunc, uint32 from, uint32 to, void * cookie) 
{ 
   uint32 size = CountItems();
   if (to > size) to = size;
   if (to > from)
   {
      if (to < from+12) 
      {
         // too easy, just do an bubble sort (base case)
         if (to > from+1) 
         {
            for (uint32 i=from+1; i<to; i++) 
            {
               for (uint32 j=i; j>from; j--) 
               {
                  int ret = compareFunc(*(GetItemPointer(j)), *(GetItemPointer(j-1)), cookie);
                  if (ret < 0) Swap(j, j-1); 
                          else break; 
               }
            } 
         } 
      }
      else
      {
         // Okay, do the real thing
         uint32 middle = (from + to)/2; 
         Sort(compareFunc, from, middle, cookie); 
         Sort(compareFunc, middle, to, cookie); 
         Merge(compareFunc, from, middle, to, middle-from, to-middle, cookie); 
      }
   }
}

template <class ItemType>
void 
PortableQueue<ItemType>::
ReverseItemOrdering(uint32 from, uint32 to) 
{
   to--;  // make it inclusive
   uint32 size = CountItems();
   if (to >= size) to = size-1;
   while (from < to) Swap(from++, to--); 
} 

template <class ItemType>
uint32 
PortableQueue<ItemType>::
RemoveAllInstancesOf(const ItemType & val) 
{
   // Efficiently collapse all non-matching slots up to the top of the list
   uint32 ret      = 0;
   uint32 writeTo  = 0;
   uint32 origSize = CountItems();
   for(uint32 readFrom=0; readFrom<origSize; readFrom++)
   {
      const ItemType & nextRead = (*this)[readFrom];
      if (nextRead == val) ret++;
      else 
      {
         if (readFrom > writeTo) (*this)[writeTo] = nextRead;
         writeTo++;
      }
   }

   // Now get rid of any now-surplus slots
   for (; writeTo<origSize; writeTo++) RemoveTail();

   return ret;
}

template <class ItemType>
void 
PortableQueue<ItemType>::
Merge(ItemCompareFunc compareFunc, uint32 from, uint32 pivot, uint32 to, uint32 len1, uint32 len2, void * cookie) 
{
   if ((len1)&&(len2))
   {
      if (len1+len2 == 2) 
      { 
         if (compareFunc(*(GetItemPointer(pivot)), *(GetItemPointer(from)), cookie) < 0) Swap(pivot, from); 
      } 
      else
      {
         uint32 first_cut, second_cut; 
         uint32 len11, len22; 
         if (len1 > len2) 
         { 
            len11      = len1/2; 
            first_cut  = from + len11; 
            second_cut = Lower(compareFunc, pivot, to, *GetItemPointer(first_cut), cookie); 
            len22      = second_cut - pivot; 
         } 
         else 
         { 
            len22      = len2/2; 
            second_cut = pivot + len22; 
            first_cut  = Upper(compareFunc, from, pivot, *GetItemPointer(second_cut), cookie); 
            len11      = first_cut - from; 
         } 

         // do a rotation
         if ((pivot!=first_cut)&&(pivot!=second_cut)) 
         {
            // find the greatest common denominator of (pivot-first_cut) and (second_cut-first_cut)
            uint32 n = pivot-first_cut;
            {
               uint32 m = second_cut-first_cut;
               while(n!=0) 
               {
                  uint32 t = m % n; 
                  m=n; 
                  n=t;
               } 
               n = m;
            }

            while(n--) 
            {
               const ItemType val = *GetItemPointer(first_cut+n); 
               uint32 shift = pivot - first_cut; 
               uint32 p1 = first_cut+n;
               uint32 p2 = p1+shift; 
               while (p2 != first_cut + n) 
               { 
                  ReplaceItemAt(p1, *GetItemPointer(p2));
                  p1 = p2; 
                  if (second_cut - p2 > shift) p2 += shift; 
                                          else p2  = first_cut + (shift - (second_cut - p2)); 
               } 
               ReplaceItemAt(p1, val);
            }
         }

         uint32 new_mid = first_cut+len22; 
         Merge(compareFunc, from,    first_cut,  new_mid, len11,        len22,        cookie); 
         Merge(compareFunc, new_mid, second_cut, to,      len1 - len11, len2 - len22, cookie); 
      }
   }
}


template <class ItemType>
uint32 
PortableQueue<ItemType>::
Lower(ItemCompareFunc compareFunc, uint32 from, uint32 to, const ItemType & val, void * cookie) const
{
   if (to > from)
   {
      uint32 len = to - from;
      while (len > 0) 
      {
         uint32 half = len/2; 
         uint32 mid  = from + half; 
         if (compareFunc(*(GetItemPointer(mid)), val, cookie) < 0) 
         {
            from = mid+1; 
            len  = len - half - 1; 
         } 
         else len = half; 
      }
   }
   return from; 
} 

template <class ItemType>
uint32 
PortableQueue<ItemType>::
Upper(ItemCompareFunc compareFunc, uint32 from, uint32 to, const ItemType & val, void * cookie) const 
{
   if (to > from)
   {
      uint32 len = to - from;
      while (len > 0) 
      { 
         uint32 half = len/2; 
         uint32 mid  = from + half; 
         if (compareFunc(val, *(GetItemPointer(mid)), cookie) < 0) len = half; 
         else 
         {
            from = mid+1; 
            len  = len - half -1; 
         } 
      } 
   }
   return from; 
}

#endif

