/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */ 

#ifndef PORTABLE_REFCOUNT_H 
#define PORTABLE_REFCOUNT_H 

#include "support/PortableTypeDefs.h" 
#include "support/PortableMacros.h" 
#include "util/ObjectPool.h" 

// Define muscle_refcount_t as the appropriate type for this configuration
#if defined(MUSCLE_SINGLE_THREAD_ONLY)
  typedef int32 muscle_refcount_t;
#elif defined(__linux__)
# include <asm/atomic.h>
  // Gotta subclass so that we can initialize it correctly... freakin Leenooks
  struct muscle_refcount_t : public atomic_t 
  {
     muscle_refcount_t() {/* empty */}
     muscle_refcount_t(int startVal) {atomic_set(this, startVal);}
  };
#elif defined(__ATHEOS__)
# include <atheos/atomic.h>
  typedef atomic_t muscle_refcount_t;
#elif defined(WIN32)
# include <windows.h>
  typedef long muscle_refcount_t;
#elif defined(__BEOS__)
# if defined(B_BEOS_VERSION_5)
  typedef vint32 muscle_refcount_t;
# else
  typedef int32 muscle_refcount_t;
# endif
#else
  typedef int32 muscle_refcount_t;
#endif

// Define the (typically atomic) add/subtract operations for this configuration
#if defined(USE_CUSTOM_ATOMIC_FUNCTIONS)
  void MUSCLE_ATOMIC_ADD(volatile muscle_refcount_t * c);      // should increment (*c)
  bool MUSCLE_ATOMIC_SUB_TEST(volatile muscle_refcount_t * c); // should decrement (*c) and return (true iff (*c) is now zero).
#elif defined(MUSCLE_SINGLE_THREAD_ONLY) 
# define MUSCLE_ATOMIC_ADD(c)      (++c)
# define MUSCLE_ATOMIC_SUB_TEST(c) (--c == 0)
#elif defined(WIN32) 
# define MUSCLE_ATOMIC_ADD(c)      (InterlockedIncrement(&c))
# define MUSCLE_ATOMIC_SUB_TEST(c) (InterlockedDecrement(&c) == 0)
#elif defined(__linux__)
# define MUSCLE_ATOMIC_ADD(c)      (atomic_inc(&c))
# define MUSCLE_ATOMIC_SUB_TEST(c) (atomic_dec_and_test(&c))
#elif defined(__ATHEOS__) 
# define MUSCLE_ATOMIC_ADD(c)      (atomic_add(&c,1))
# define MUSCLE_ATOMIC_SUB_TEST(c) (atomic_add(&c,-1)==1)
#elif defined(__BEOS__) 
# define MUSCLE_ATOMIC_ADD(c)      (atomic_add(&c,1)) 
# define MUSCLE_ATOMIC_SUB_TEST(c) (atomic_add(&c,-1)==1)
#else
# error "No atomic inc/dec calls are supplied for this OS!  Add your own here in PortableRefCount.h, or put -DMUSCLE_SINGLE_THREAD_ONLY in your Makefile if you won't be using multithreading, or define USE_CUSTOM_ATOMIC_FUNCTIONS and supply your own atomic functions in your source code." 
#endif 


/** 
 *  If you derive your own classes from this class, PortableRef can
 *  handle them more efficiently.  PortableRef can also handle types that
 *  are not derived from this class, however.
 */
class PortableRefCountable
{
public:
   /** Default constructor.  Refcount begins at zero. */
   PortableRefCountable() : _refCount(0) {/* empty */}

   /** Copy constructor -- does NOT copy the refcount! */
   PortableRefCountable(const PortableRefCountable &) : _refCount(0) {/* empty */}

   /** Assigment operator.  Implemented as a no-op; does NOT copy the refcount! */
   PortableRefCountable &operator=(const PortableRefCountable &) {return *this;}

   /** Increments the counter */
   void IncrementRefCount() const {MUSCLE_ATOMIC_ADD(_refCount);}

   /** Decrements the counter and returns true iff it has reached zero. */
   bool DecrementRefCount() const {return MUSCLE_ATOMIC_SUB_TEST(_refCount);}

private:
   mutable muscle_refcount_t _refCount;
};

/**
 *  This is a ref-count token object that works with any instantiation
 *  type that is a subclass of PortableRefCountable.  The old implementation
 *  of PortableRef, which could use any C++ type at all, has been renamed
 *  to UnintrusivePortableRef.
 */
template <class Item> class MUSCLE_CLASS_DECLARATIONS PortableRef
{
public:
   typedef ObjectPool<Item> ItemPool;        /**< type of an ObjectPool of user data structures */

   /** 
    * Creates a new reference-count for the given item.
    * Once this reference count has been created, (item) will be automatically
    * deleted when the last PortableRef that references it has dissappeared.
    * Note that it is an error to call this constructor twice with the same (item).
    * To obtain additional tracked references to (item), use PortableRef's copy 
    * constructor, or the PortableRef equals operator.
    *
    * (itemPool) may be NULL, or if you wish to be more efficient, it may point to a ItemPool
    * that you are using to recycle your (item)s.  (Note that
    * the calling code retains ownership of the ItemPool object)
    *
    * Usage example: 
    *
    * class Foo : public PortableRefCountable {...};
    * PortableRef<Foo> ref(new Foo(), NULL);
    * for (int i=0; i<10; i++) _vectors[i].AddItem(ref);
    * (the Foo object will be deleted automagically at the right time, later...)
    *
    * @param item A dynamically allocated object that the PortableRef class will assume responsibility for deleting.
    * @param itemPool An optional ObjectPool to recycle the object into.  May be NULL.
    */
   PortableRef(Item * item, ItemPool * itemPool)
   {
      _item       = item;
      _itemPool   = itemPool;
      _numItems   = 1;
      _deleteItem = true;
      RefItem();
   }
 
   /** 
    * Creates a new reference-count for the given item or item array.
    * Once this reference count has been created, (item) will be automatically
    * deleted when the last PortableRef that references it has dissappeared.
    * Note that it is an error to call this constructor twice with the same (item).
    * To obtain additional tracked references to (item), use PortableRef's copy 
    * constructor, or the PortableRef equals operator.
    *
    * (itemPool) may be NULL, or if you wish to be more efficient, it may point to a ItemPool
    * that you are using to recycle your (item)s.  (Note that
    * the calling code retains ownership of the ItemPool object)
    *
    * If (item) is a pointer to a single item, and not an array of items, then
    * the default value of one is appropriate.
    *
    * (numItems) is the number of (Item) objects that (item) points to.  In general,
    * this will be one, but for arrays it might be more.
    *
    * If (doRefCounting) is set false, then (item) will NOT be deleted automatically.
    * It then remains the responsibility of the calling code to delete it.  (This is useful
    * if e.g. you need to put a non-new'd item into a PortableRef, and want to 
    * handle the cleanup issues yourself)
    *
    * Usage example: 
    *
    * class Foo : public PortableRefCountable {...};
    * PortableRef<Foo> ref(new Foo(), NULL, 5, true);
    * for (int i=0; i<10; i++) _vectors[i].AddItem(ref);
    * (the Foo object will be deleted automagically at the right time, later...)
    *
    * @param item A dynamically allocated object that the PortableRef class will assume responsibility for deleting.
    * @param itemPool An optional ObjectPool to recycle the object into.  May be NULL.
    * @param numItems Number of Items pointed to by (item).  Usually 1, but may be different if (item)
    *                 is actually pointing to an array of Items.
    * @param doRefCounting If set false, the PortableRef will never delete the Item (or array).  
    *                      (this is only useful in very rare cases)  If set to false, the calling
    *                      code will retain responsibility for deleting Item at the proper time.
    */
   PortableRef(Item * item, ItemPool * itemPool, uint32 numItems, bool doRefCounting)
   {
      _item       = item;
      _itemPool   = itemPool;
      _numItems   = numItems;
      _deleteItem = doRefCounting;
      RefItem();
   }
 

   /** Copy constructor.  Creates an additional reference to the object referenced by (copyMe).
    *  The referenced object won't be deleted until ALL PortableRefs that reference it are gone.
    */
   PortableRef(const PortableRef & copyMe) : _item(NULL)
   {
      *this = copyMe;
   }

   /** Default constructor.
    *  Creates a NULL reference (suitable for initializing with the = operator)
    */
   PortableRef() : _item(NULL)
   {
      // empty
   }

   /** Unreferences the held data item.  If this is the last PortableRef that
    *  references the held data item, the data item will be deleted or recycled at this time.
    */
   ~PortableRef() 
   {
      UnrefItem();
   }

   /** Assigment operator.
    *  Unreferences the previous held data item, and adds a reference to the the data item of (r).
    */
   PortableRef &operator=(const PortableRef &r) 
   {
      if (&r != this)
      {
         UnrefItem();
         _item       = r._item;
         _itemPool   = r._itemPool;
         _numItems   = r._numItems;
         _deleteItem = r._deleteItem;
         RefItem();
      }
      return *this;
   }

   /** Returns true iff both PortableRefs are referencing the same data. */
   int operator ==(const PortableRef &rhs) const {return _item == rhs._item;}
 
   /** Returns the ref-counted data item.  The returned data item
    *  is only guaranteed valid for as long as this PortableRefCount object exists.
    */
   Item * GetItemPointer() const {return _item;}

   /** Convenience synonym for GetItemPointer(). */
   Item * operator()() const {return GetItemPointer();}

   /** Returns the number of items in the array we are tracking (or
    *  one if we are tracking just a single object)
    */
   uint32 GetItemCount() const {return _itemCount;}

   /** Unreferences our held data item (if any), and turns this object back into a NULL reference.  
    *  (equivalent to *this = PortableRef();)
    */
   void Reset() {UnrefItem();}

private:
   void RefItem() {if (_item) _item->IncrementRefCount();}
   void UnrefItem()
   {
      if (_item)
      {
         if ((_item->DecrementRefCount())&&(_deleteItem))
         {
            if (_itemPool) _itemPool->ReleaseObject(_item, _numItems);
            else
            {
               if (_numItems == 1) delete _item;
                              else delete [] _item;
            }
         }
         _item = NULL;
      }
   }
   
   Item * _item; 
   ItemPool * _itemPool;
   uint32 _numItems;
   bool _deleteItem;
};

/**
 *  This class implements reference counting on any object or array
 *  you give it, so that your object will be automatically deleted when 
 *  your last reference to it goes away.  This one works on any C++ type,
 *  and not just subclasses of PortableRefCountable; but it is somewhat
 *  less efficient than the PortableRef class.
 */
template <class Item> class MUSCLE_CLASS_DECLARATIONS UnintrusivePortableRef
{
private:
   class RefCountMem;

public:
   typedef ObjectPool<RefCountMem> RefPool;  /**< type of an ObjectPool of our internal data structures */
   typedef ObjectPool<Item> ItemPool;        /**< type of an ObjectPool of user data structures */

   /** 
    * Creates a new reference-count for the given item.
    * Once this reference count has been created, (item) will be automatically
    * deleted when the last UnintrusivePortableRef that references it has dissappeared.
    * Note that it is an error to call this constructor twice with the same (item).
    * To obtain additional tracked references to (item), use UnintrusivePortableRef's copy 
    * constructor, or the UnintrusivePortableRef equals operator.
    *
    * (refPool) may be NULL, or if you wish to be more efficient, it may point to a RefPool
    * that you are using to recycle ref-count memory for this UnintrusivePortableRef class.  (Note that
    * the calling code retains ownership of the RefPool object)
    *
    * (itemPool) may be NULL, or if you wish to be more efficient, it may point to a ItemPool
    * that you are using to recycle your (item)s.  (Note that
    * the calling code retains ownership of the ItemPool object)
    *
    * If (item) is a pointer to a single item, and not an array of items, then
    * the default value of one is appropriate.
    *
    * (numItems) is the number of (Item) objects that (item) points to.  In general,
    * this will be one, but for arrays it might be more.
    *
    * If (doRefCounting) is set false, then (item) will NOT be deleted automatically.
    * It then remains the responsibility of the calling code to delete it.  (This is useful
    * if e.g. you need to put a non-new'd item into a UnintrusivePortableRef, and want to 
    * handle the cleanup issues yourself)
    *
    * Usage example: 
    *
    * UnintrusivePortableRef<Foo> ref(new Foo());
    * for (int i=0; i<10; i++) _vectors[i].AddItem(ref);
    * (the Foo object will be deleted automagically at the right time, later...)
    *
    * @param item A dynamically allocated object that the UnintrusivePortableRef class will assume responsibility for deleting.
    * @param itemPool An optional ObjectPool to recycle the object into.  May be NULL.
    * @param refPool An optional ObjectPool to recycle our internal ref-count data into.  May be NULL.
    * @param numItems Number of Items pointed to by (item).  Usually 1, but may be different if (item)
    *                 is actually pointing to an array of Items.
    * @param doRefCounting If set false, the UnintrusivePortableRef will never delete the Item.  
    *                      (this is only useful in very rare cases)  If set to false, the calling
    *                      code will retain responsibility for deleting Item at the proper time.
    */
   UnintrusivePortableRef(Item * item, ItemPool * itemPool, RefPool * refPool, uint32 numItems = 1, bool doRefCounting = true)
   {
      if (doRefCounting)
      {
         _rc = (refPool) ? refPool->GetObject() : newnothrow RefCountMem();
         if (_rc) 
         {
            _rc->Init(item, itemPool, refPool, numItems);
            _defaultPtr = NULL;
            _ownedItemCount = -1;  // _rc owns the pointer now
         }
         else
         {
            // can't allocate refCount!  Fall back to just holding it ourself (sole ownership)
            WARN_OUT_OF_MEMORY;
            _defaultPtr = item;
            _ownedItemCount = numItems;  // we own the pointer
         }
      }
      else
      {
         _rc = NULL;
         _defaultPtr = item;
         _ownedItemCount = -1;  // calling code still owns the pointer here
      }
   }

   /** Copy constructor.  Creates an additional reference to the object referenced by (copyMe).
    *  The referenced object won't be deleted until ALL UnintrusivePortableRefs that reference it are gone.
    */
   UnintrusivePortableRef(const UnintrusivePortableRef & copyMe) : _rc(NULL), _defaultPtr(NULL), _ownedItemCount(-1)
   {
      *this = copyMe; 
   }

   /** Default constructor.
    *  Creates a NULL reference (suitable for initializing with the = operator)
    */
   UnintrusivePortableRef() : _rc(NULL), _defaultPtr(NULL), _ownedItemCount(-1) 
   {
      // empty
   }

   /** Unreferences the held data item.  If this is the last UnintrusivePortableRef that
    *  references the held data item, the data item will be deleted or recycled at this time.
    */
   ~UnintrusivePortableRef() 
   {
      Reset();
   }

   /** Assigment operator.
    *  Unreferences the previous held data item, and adds a reference to the the data item of (r).
    */
   UnintrusivePortableRef &operator=(const UnintrusivePortableRef &r) 
   {
      if (&r != this)
      {
         DeleteOwnedDefaultPointer();
         if (_rc) _rc->Unref();
         _rc = r._rc;
         if (_rc) _rc->Ref();
         _defaultPtr = (r._ownedItemCount >= 0) ? NULL : r._defaultPtr;  // if he is sole owner, we don't want it
      }
      return *this;
   }

   /** Returns true iff both UnintrusivePortableRefs are referencing the same data. */
   int operator ==(const UnintrusivePortableRef &rhs) const {return _rc == rhs._rc;}
 
   /** Returns the ref-counted data item.  The returned data item
    *  is only guaranteed valid for as long as this UnintrusivePortableRefCount object exists.
    */
   Item * GetItemPointer() const {return _rc ? _rc->GetItem() : _defaultPtr;}

   /** Convenience synonym for GetItemPointer(). */
   Item * operator()() const {return GetItemPointer();}

   /** Returns the number of items in the array we are tracking (or
    *  one if we are tracking just a single object)
    */
   uint32 GetItemCount() const {return _rc ? _rc->GetItemCount() : 0L;}

   /** Convenience method:  Unreferences our held data item (if any), and 
    *  turns this object back into a NULL reference.  (equivalent to *this = UnintrusivePortableRef();)
    */
   void Reset() 
   {
      if (_rc) _rc->Unref();
      _rc = NULL;
      DeleteOwnedDefaultPointer();
   }
   
private:
   void DeleteOwnedDefaultPointer()
   {
      if (_ownedItemCount >= 0) 
      {
         if (_ownedItemCount == 1) delete _defaultPtr;
                              else delete [] _defaultPtr;
         _ownedItemCount = -1;
         _defaultPtr = NULL;
      }
   }

   /** This class is used internally by the UnintrusivePortableRef class.  An object of this class
    *  represents the reference count for a given object, and is shared by all the UnintrusivePortableRef
    *  classes referencing the tracked object.
    */
   class RefCountMem
   { 
   public:
      RefCountMem() {/* empty */}

      void Init(Item * item, ItemPool * itemPool, RefPool * refPool, uint32 numItems)
      {
         _refCount  = 1;
         _item      = item;
         _itemPool  = itemPool;
         _refPool   = refPool;
         _numItems  = numItems;
      }

      ~RefCountMem() 
      {
         FreeObject();  // shouldn't be necessary but just in case
      }

      void FreeObject()
      {
         if (_item)
         {
            if (_itemPool) _itemPool->ReleaseObject(_item, _numItems);
            else 
            {
               if (_numItems == 1) delete _item;
                              else delete [] _item;
            }
            _item = NULL;
         }
      }
  
      void Ref()  
      {
         MUSCLE_ATOMIC_ADD(_refCount); 
      }

      void Unref() 
      {
         if (MUSCLE_ATOMIC_SUB_TEST(_refCount))
         {
            FreeObject();
            if (_refPool) _refPool->ReleaseObject(this);
                     else delete this;
         }
      }

      Item * GetItem() const {return _item;}

   private:
      Item * _item;
      muscle_refcount_t _refCount;
      RefPool * _refPool;
      ItemPool * _itemPool;
      uint32 _numItems;
   };

   RefCountMem * _rc;
   Item * _defaultPtr;  // used only when user is handling object deletion himself
   int _ownedItemCount;  // if >=0, we have sole ownership of (_defaultPtr), and _ownedItemCount is the array size
};

#endif
